/* --------------------------------------------------------------- */
/* ADXL345.c -- accelerometer device I/O for Tollos                */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* These functions provide basic IO for using the ADXL345          */
/* accelerometer device, attached via I2C.                         */
/*                                                                 */
/* The following connections are assumed:                          */
/*                                                                 */
/* * SCL & SDA ... to I2C bus SCL & SDA, with appropriate pullups  */
/* (channel is selectable)                                         */
/* * SDO ... to GND (selects device alternate physical bus address)*/
/* * CS ... to VCC (+3.3V) [enable I2C mode]                       */
/* * INT1 ... to a suitable input bit pin                          */
/* * VCC & GND ... to suitable power rails, with bypass capacitors */
/*                                                                 */
/* (INT2 is currently unused.)                                     */
/*                                                                 */
/* This code assumes only one ADXL345 device will be used on a     */
/* controller, so some state information can be held (the device   */
/* channel and address).  This simplifies the tollosAcc interface  */
/* and allows initialization checking.                             */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO       // to check signatures only
#include "tollos.h"
#include "XL345.h"       // Analog Devices' XL345.h

// state
static uint accAddr=0;   // set to i2c device address when initialized

/* --------------------------------------------------------------- */
/* accGetXYZIO -- read latest x, y, and z accelerations            */
/*                                                                 */
/*   x, y, z -- receive accelerations [in milli-gs]                */
/*                                                                 */
/*   returns 0 if data read, -1 if the device is not initialized,  */
/*     or -2 if there was an error reading the data (or no data    */
/*     ready); 1 is returned if data overrun was raised -- in this */
/*     case data is read                                           */
/*                                                                 */
/* Results are returned scaled to milli-gs (1000 = 1g), without    */
/* any calibration adjustments.                                    */
/*                                                                 */
/* This should normally be called only after a Data Ready          */
/* interrupt; the interrupt is cleared by calling this.            */
/* --------------------------------------------------------------- */
int accGetXYZIO(int *x, int *y, int *z) {
  byte status;           // work
  int  rc;               // ..
  int16 value;           // ..
  byte data[6];          // read buffer X0, X1, Y0, Y1, Z0, Z1

  if (accAddr==0) return -1;  // stopped

  // Reading the INT_SOURCE register should clear any interrupts,
  // according to the datasheet, but in fact DATA_READY is cleared
  // by the Read of DATAX0.
  //
  // DATA_READY, Overrun, and Watermark are set even if not enabled
  // for interrupt.
  rc=i2cGetReg(accAddr, XL345_INT_SOURCE, &status);

  // always read the data registers to clear overrun, etc.
  rc|=i2cGetRegs(accAddr, XL345_DATAX0, data, 6);

  // The time from INT1 raised until all 6 bytes are read is 0.332ms,
  // when using i2c at nominally 400 kHz (actual bus clock = 375khz
  // on LPC1768 with 96Mhz clock set)

  if (!(status&XL345_DATAREADY)) return -2;

  if (rc!=0) return -3;            // i2c error

  // one ulp is nominally 4 milli-g, but from the datasheet details
  // it seems that it is actually 1000/256 mg (3.90625)
  // [16g does not overflow 16 bits]
  value=(data[1]<<8) | data[0];    // assemble 16 bits
  *x=(1000*(int)value/256);        // [propagates sign]
  value=(data[3]<<8) | data[2];
  *y=(1000*(int)value/256);
  value=(data[5]<<8) | data[4];
  *z=(1000*(int)value/256);

  if (status&XL345_OVERRUN) return 1;
  return 0;
  } // accGetXYZIO

/* --------------------------------------------------------------- */
/* accInitIO -- configure and initialize accelerometer             */
/*                                                                 */
/*   channel -- I2C channel on which device is to be found         */
/*                                                                 */
/*   returns 0 if initialized OK, negative if error                */
/*                                                                 */
/* The tollosAcc interface expects:                                */
/* * an update rate of at least 25 Hz (but need not be higher than */
/* that on initialization, to save power)                          */
/* * a range of at least 2g                                        */
/* * an interrupt [from pin INT1] when Data Ready is raised.       */
/*                                                                 */
/* The caller should set up the interrupt handler before calling   */
/* this function, or clear Data Ready by calling accGetXYZIO.      */
/* --------------------------------------------------------------- */
int accInitIO(uint channel) {
  byte b;                // work
  int  rc;               // ..
  byte data[6];          // read buffer

  accAddr=i2cDevice(channel, XL345_ALT_ADDR);

  // check the right chip is there
  rc=i2cGetReg(accAddr, XL345_DEVID, &b);
  if (rc!=0 || b!=XL345_ID) return -1;            // hopeless

  // from here, rc is accumulated by ORing, so any negative rc
  // will result in a final negative rc

  // disable interrupts and stop device (enter standby mode)
  rc|=i2cPutReg(accAddr, XL345_INT_ENABLE, 0x00);
  rc|=i2cPutReg(accAddr, XL345_POWER_CTL, 0x00);

  // set +/- 2g, full resolution, normal justification, interrupt
  // active high
  rc|=accRangeIO(2000);

  // set data rate 25 Hz, and request low noise
  // [LOW_POWER mode would drop 60uA -> 40uA at 25Hz and 2.5V]
  rc|=i2cPutReg(accAddr, XL345_BW_RATE, XL345_RATE_25 | XL345_LOW_NOISE);

  // map all interrupts to INT1 (set bits to 0) and enable DATA_READY
  rc|=i2cPutReg(accAddr, XL345_INT_MAP, 0x00);
  rc|=i2cPutReg(accAddr, XL345_INT_ENABLE, XL345_DATAREADY);

  // read data then status to clear Data Ready and Overrun
  rc|=i2cGetRegs(accAddr, XL345_DATAX0, data, 6);
  rc|=i2cGetReg(accAddr, XL345_INT_SOURCE, &b);

  // start measuring
  rc|=i2cPutReg(accAddr, XL345_POWER_CTL, XL345_MEASURE);

  return rc;
  } // accInitIO

/* --------------------------------------------------------------- */
/* accRangeIO -- change range of accelerometer                     */
/*                                                                 */
/*   range -- range requested in milli-g [1000 = 1 g]              */
/*                                                                 */
/*   returns 0 if range changed OK, -1 if error (range too high)   */
/*                                                                 */
/* The accelerometer will be set to a smallest available range     */
/* that is >= the requested range, to maximize the precision of    */
/* returned results.                                               */
/* --------------------------------------------------------------- */
int accRangeIO(uint range) {
  byte gset;             // setting to use

  if (range<=2000)        gset=XL345_RANGE_2G;
   else if (range<=4000)  gset=XL345_RANGE_4G;
   else if (range<=8000)  gset=XL345_RANGE_8G;
   else if (range<=16000) gset=XL345_RANGE_16G;
   else return -1;

  // set g, full resolution, normal justification, and interrupt
  // active high
  return i2cPutReg(accAddr, XL345_DATA_FORMAT,
                   gset
                 | XL345_FULL_RESOLUTION
                 | XL345_DATA_JUST_RIGHT
                 | XL345_INT_HIGH);
  } // accRangeIO

/* --------------------------------------------------------------- */
/* accRateIO -- change update rate of accelerometer                */
/*                                                                 */
/*   rate -- rate requested [Hz, 1..200]                           */
/*                                                                 */
/*   returns 0 if rate changed OK, -1 if error (rate zero or       */
/*     too high)                                                   */
/*                                                                 */
/* The accelerometer will be set to a slowest available rate       */
/* that is >= the requested rate, to minimize resource usage.      */
/*                                                                 */
/* To ensure working with 100kHz i2c, the maximum rate is limited  */
/* to 200Hz -- but note that this may use significant processor    */
/* bandwidth.                                                      */
/* --------------------------------------------------------------- */
int accRateIO(uint rate) {
  byte gset;             // setting to use

  if (rate==0) return -1;
   else if (rate==1)   gset=XL345_RATE_1_563;
   else if (rate<=3)   gset=XL345_RATE_3_125;
   else if (rate<=6)   gset=XL345_RATE_6_25;
   else if (rate<=12)  gset=XL345_RATE_12_5;
   else if (rate<=25)  gset=XL345_RATE_25;
   else if (rate<=50)  gset=XL345_RATE_50;
   else if (rate<=100) gset=XL345_RATE_100;
   else if (rate<=200) gset=XL345_RATE_200;
   else return -1;

  // set data rate, with low noise
  // [LOW_POWER mode would drop 60uA -> 40uA at 25Hz and 2.5V]
  return i2cPutReg(accAddr, XL345_BW_RATE, gset | XL345_LOW_NOISE);
  } // accRateIO

/* --------------------------------------------------------------- */
/* accStopIO -- stop and power-down accelerometer                  */
/*                                                                 */
/*   returns 0 if accelerometer stopped OK, negative if an error   */
/*     or it was not running.                                      */
/* --------------------------------------------------------------- */
int accStopIO(void) {
  uint da=accAddr;            // save
  if (accAddr==0) return -1;  // already stopped
  accAddr=0;                  // always require re-initialize

  // disable interrupts and stop device (enter standby mode)
  i2cPutReg(accAddr, XL345_INT_ENABLE, 0x00); // don't care if this fails
  return i2cPutReg(da, XL345_POWER_CTL, 0x00);
  } // accStopIO

